	/*
            ARTHACHITRA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE CODE FOR ANY 
            PURPOSE. IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY OF ANY KIND. ARTHACHITRA DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOURCE CODE, 
            INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY, NON-INFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL ARTHACHITRA BE LIABLE FOR ANY 
            SPECIAL, INDIRECT, INCEDENTAL, OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION, ARISING 
            OUT OF OR IN CONNECTION WITH THE USE OR PERFORMACE OF THIS SOURCE CODE.
    */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using SharpCharts.Base.Data;
using SharpCharts.Base.Order;
using System.Diagnostics;
using System.Globalization;

namespace SharpCharts.Base.Connection
{

	internal enum ColumnType
	{
		BidRate,
		AskRate,
		LastTradedPrice,
		BidQuantity,
		AskQuantity,
		TotalVolume
	}
	
		
	internal static class NestExchange
	{
		public static string NSE
		{
			get { return "nse_cm"; }
		}
		
		public static string NSEFO
		{
			get { return "nse_fo"; }
		}
		
		public static string BSE
		{
			get { return "bse_cm"; }
		}
		
		public static string BSEFO
		{
			get { return "bse_fo"; }
		}
		
		public static string MCXFO
		{
			get { return "mcx_fo"; }
		}
		
		public static string NSECDS
		{
			get { return "cde_fo"; }
		}
	}
	
	
    public class NestTraderConnection : ExternalConnectionBase, NestTrader.IRTDUpdateEvent
    {
    
    	private class RTDQuotes
    	{
    		double bidPrice, askPrice, lastPrice;
    		int bidQty, askQty, totalVol;
    		bool canAppendBid, canAppendAsk, isInitialized;
    		
    		public RTDQuotes(Instrument instrument)
    		{
    			this.Instrument = instrument;
    		}
    		
    		public bool CanAppendQuotes(ColumnType columnType, double value)
    		{
    			switch (columnType)
        		{
        			case ColumnType.BidRate:
        			
        				if (this.bidPrice != value)
        				{
        					this.canAppendBid = true;
        					this.bidPrice = value;
        				}
        				break;
					case ColumnType.AskRate:
						if (this.askPrice != value)
						{
							this.canAppendAsk = true;
							this.askPrice = value;
						}
						break;
					case ColumnType.LastTradedPrice:
						this.lastPrice = value;
						break;
				}
    			
    			return false;
    		}
    		
    		
    		public bool CanAppendQuotes(ColumnType columnType, int value, out double price, out int volume)
    		{
    		
    			price = 0.0d;
    			volume = 0;
    			
    			switch (columnType)
    			{
    				case ColumnType.BidQuantity:    					
						if (this.bidQty != value)
						{
							this.bidQty = value;
							this.canAppendBid = true;
						}
						
						if (this.canAppendBid)
						{
							price = this.bidPrice;
							volume = value;
							this.canAppendBid = false; //reset it
							return true;
						}
						break;
					case ColumnType.AskQuantity:
						if (this.askQty != value)
						{
							this.askQty = value;
							this.canAppendAsk = true;
						}
						
						if (this.canAppendAsk)
						{
							price = this.askPrice;
							volume = value;
							this.canAppendAsk = false;
							return true;
						}
						
						break;
					case ColumnType.TotalVolume:
					
						if (value == 0)
						{
							this.totalVol = 0;
							this.isInitialized = true;
							return false;
						}
						
						if (!this.isInitialized)
						{
							this.isInitialized = true;
							this.totalVol = value;
							return false;
						}
					
						if (value > this.totalVol)
						{
							price = this.lastPrice;
							volume = value - this.totalVol;
							this.totalVol = value;
							
							return true;
						}
						break;
    			}
    			
    			return false;
    		}
    		
    		public Instrument Instrument { get; private set; }
    	}
    	
    	private class RTDData
    	{
    		
    		public RTDData(ColumnType columnType, RTDQuotes rtdQutoes)
    		{
    			this.ColumnType = columnType;
    			this.RTDQuote = rtdQutoes;
    		}
    		
    		public ColumnType ColumnType { get; private set; }
    		public RTDQuotes RTDQuote { get; private set; }
    	}
    
    
    	private object locker = new object();
    	
    	List<Instrument> instruments = null;
    	Dictionary<int, RTDData> quotes = null;
        
        NestTrader.ScripRTDClass rtd = null;
        Process process = null;
        TimeSpan timezoneOffset = TimeSpan.Zero;
        
        int topicId;
        
        DateTime lastUpdateTime;
        bool isUpdating;
        
        public override void Connect(IConnection connection)
        {
            if (!Common.Globals.IsAdministrator)
            {
                Log("You must run the application as an Administrator", logCategory: Base.Log.LogCategory.Connection);
                SetConnectionState(ConnectionState.ConnectionFailed);
                return;
            }
            
            this.process = Process.GetProcessesByName("NestTrader").FirstOrDefault();

            if (this.process == null)
            {
                Log("NestTrader not running", logCategory: Base.Log.LogCategory.Connection);
                SetConnectionState(ConnectionState.ConnectionFailed);
                return;
            }
            
    		timezoneOffset = Common.Globals.GetUTCTimeZoneOffset(Common.TimeZone.IndiaStandardTime);

			this.instruments = new List<Instrument>();
			this.quotes = new Dictionary<int, NestTraderConnection.RTDData>();

            try
            {
                this.rtd = new NestTrader.ScripRTDClass();
            	this.rtd.ServerStart(this);
                System.Threading.Thread.Sleep(1000);
            }
            catch (Exception ex)
            {
                this.rtd = null;
                Log(ex.Message);
                SetConnectionState(ConnectionState.ConnectionFailed);
                return;
            }

            this.topicId = 1;
            this.isUpdating = false;
            this.lastUpdateTime = DateTime.MinValue;

            this.process.Exited += OnNestTraderExited;

            SetConnectionState(ConnectionState.Connected);
        }
        
        public override void Disconnect()
        {
        	Disconnect(ConnectionState.Disconnected);    
        }
        
        private void Disconnect(ConnectionState connectionState)
        {
        	Process tmpProcess = this.process;
        	if (tmpProcess != null)
        	{
        		tmpProcess.Exited -= OnNestTraderExited;
        	}
			
			NestTrader.ScripRTDClass tmpRtd = this.rtd;
			if (tmpRtd != null)
			{
			
				Dictionary<int, RTDData> tmpQuotes = this.quotes;
				if (tmpQuotes != null)
				{
					int[] keys = tmpQuotes.Keys.ToArray();
					for (int i = 0; i < keys.Length; i++)
					{
						try	
						{
							tmpRtd.DisconnectData(keys[i]);
						}
						catch (Exception ex)
						{
							Log(ex.Message);
						}
					}
				}
				
				this.quotes = null;
				this.instruments = null;
				this.topicId = 1;
				
				try	
				{
					tmpRtd.ServerTerminate();
				}
				catch (Exception ex)
				{
					Log(ex.Message);
				}
			}
			
			SetConnectionState(connectionState);
        }
        
        private void OnNestTraderExited(object sender, EventArgs args)
        {
        	Log("Nest Trader exited");
        	this.Disconnect();
        }
        
        private string GetColumn(ColumnType columnType)
        {
        	switch (columnType)
        	{
        		case ColumnType.BidRate:
					return "Bid Rate"; 
				case ColumnType.AskRate:
					return "Ask Rate";
				case ColumnType.LastTradedPrice:
					return "LTP"; 
				case ColumnType.BidQuantity:
					return "Bid Qty";
				case ColumnType.AskQuantity:
					return "Ask Qty";
				case ColumnType.TotalVolume:
					return "Volume Traded Today";
				default:
					return "LTP";
			}
		}
		
		
        
        private string GetExchange(Instrument instrument)
        {
        	if (instrument == null || instrument.InstrumentDefination == null)
        		return string.Empty;
        	
        	InstrumentType instrumentType = instrument.InstrumentDefination.InstrumentType;
        	Exchange exchange = instrument.InstrumentDefination.Exchange;
        	
        	switch (instrumentType)
        	{
        		case InstrumentType.Stocks:
        			if (exchange == Exchange.NSE)
        				return NestExchange.NSE;
        			else if (exchange == Exchange.BSE)
        				return NestExchange.BSE;
        			break;
        		case InstrumentType.Futures:
        		case InstrumentType.Options:
        			if (exchange == Exchange.NSE)
        				return NestExchange.NSEFO;
        			else if (exchange == Exchange.BSE)
        				return NestExchange.BSEFO;
        			else if (exchange == Exchange.MCX)
        				return NestExchange.MCXFO;
        			else if (exchange == Exchange.NSECDS)
        				return NestExchange.NSECDS;
        			break;
        	}
        	
        	return string.Empty;
        }
        
        private string GetSymbol(Instrument instrument)
        {
        	if (instrument == null || instrument.InstrumentDefination == null)
        		return string.Empty;
        
        	string symbol = GetSymbolByDataFeedProvider(instrument);
        	if (string.IsNullOrEmpty(symbol))
        		return string.Empty;
        		
        	if (instrument.InstrumentDefination.InstrumentType == InstrumentType.Futures)
			{
				return string.Format(CultureInfo.InvariantCulture, "{0}{1:yyMMM}FUT", symbol, instrument.ExpiryDate).ToUpper();
			}
			else if (instrument.InstrumentDefination.InstrumentType == InstrumentType.Options)
        	{
        		string optionType = instrument.OptionType == OptionType.Put ? "PE" : "CE";
        	
        		if (instrument.InstrumentDefination.IsWeeklyContract)
        		{
        			return string.Format(CultureInfo.InvariantCulture, "{0}{1:ddMMMyy}{2}{3}", symbol, instrument.ExpiryDate, instrument.StrikePrice, optionType).ToUpper();
        		}
        		else
        		{
        			return string.Format(CultureInfo.InvariantCulture, "{0}{1:yyMMM}{2}{3}", symbol, instrument.ExpiryDate, instrument.StrikePrice, optionType).ToUpper();
        		}
        	}
        	
        	//<SYMBOL>|<EXCHANGE>|<SECTYPE>|<TRADING CLASS>|<MULTIPLIER>
        	//For series other than EQ, please specify the full symbol in the trading class field
        	string[] array = symbol.Split('|');
        	
        	if (array.Length == 0)
        		return string.Empty;
        	
        	if (array.Length >= 4 && !string.IsNullOrWhiteSpace(array[3]))
        	{
        		return array[3].Trim().ToUpper();
        	}
        	
        	if (string.IsNullOrWhiteSpace(array[0]))
        		return string.Empty;
        		
        	return array[0].Trim().ToUpper() + "-EQ";      	
        }
        
        
        

        public override async Task SubscribeLevel1(Instrument instrument)
        {
        	if (instrument == null || instrument.InstrumentDefination == null)
        		return;
        
			if (rtd == null)
				return;
			
			
			string exchange = GetExchange(instrument);
			if (string.IsNullOrEmpty(exchange))
			{
				Log(string.Format("Unknown exchange for instrument {0}", instrument.DisplayName), logCategory: Base.Log.LogCategory.Connection);
				return;
			}
			
			string symbol = GetSymbol(instrument);
			if (string.IsNullOrEmpty(symbol))
			{
				Log(string.Format("No symbol found for instrument {0}", instrument.DisplayName), logCategory: Base.Log.LogCategory.Connection);
				return;
			}
			
			lock (locker)
			{
				if (instruments.Contains(instrument))
					return;
					
				instruments.Add(instrument);
			}	
			
			RTDQuotes quote = new NestTraderConnection.RTDQuotes(instrument);
			
			ColumnType[] columns = (ColumnType[])Enum.GetValues(typeof(ColumnType));
			
			for (int i = 0; i < columns.Length; i++)
			{
				Array obj = new object[2];
				obj.SetValue(string.Format("{0}|{1}", exchange, symbol), 0);
				obj.SetValue(GetColumn(columns[i]), 1);

				bool newValue = true;
				
				int tmpTopic = this.topicId++;

				try
				{
					
					object val = rtd.ConnectData(tmpTopic, ref obj, ref newValue);
					if (val != null)
					{
						this.quotes.Add(tmpTopic, new RTDData(columns[i], quote));
					}
				}
				catch (Exception ex)
				{
					Log(ex.Message, logCategory: Base.Log.LogCategory.Connection);
					return;
				}
				
			}
        }


        public override bool IsOrderFeed
        {
            get { return false; }
        }


        public override async Task SubmitOrder(IOrder order)
        {
            
        }

        public override async Task ChangeOrder(IOrder order)
        {

        }

        public override async Task CancelOrder(IOrder order)
        {

        }
        
        
        #region IRTUpdateEvents
        
        void NestTrader.IRTDUpdateEvent.Disconnect()
		{
			if (base.ConnectionState == ConnectionState.Connected)
			{
				Log("Disconnected by Nest Trader", logCategory: Base.Log.LogCategory.Connection);
				Disconnect(ConnectionState.Disconnected);
			}
		}
		
		void NestTrader.IRTDUpdateEvent.UpdateNotify()
		{
			NestTrader.ScripRTDClass rtd = this.rtd;
			if (rtd == null)
				return;
				
			Dictionary<int, RTDData> tmpQuotes = this.quotes;
			if (tmpQuotes == null)
				return;			
			
			if (rtd.Heartbeat() <= 0)
				return;

			int tmp = -1;
			Array array = rtd.RefreshData(ref tmp);
			
			if (array == null)
				return;
			
			DateTime time = DateTime.UtcNow.Add(this.timezoneOffset);
			
			for	(int i = 0; i < array.GetLength(1); i++)
			{
				int tmpTopicId = (int)array.GetValue(0, i);
				
				RTDData data = null;
				if (tmpQuotes.TryGetValue(tmpTopicId, out data))
				{
					if (array.GetValue(1, i) is double)
					{
						data.RTDQuote.CanAppendQuotes(data.ColumnType, (double)array.GetValue(1, i));
					}
					else if (array.GetValue(1, i) is int)
					{
						double price;
						int volume;
						if (data.RTDQuote.CanAppendQuotes(data.ColumnType, (int)array.GetValue(1, i), out price, out volume))
						{
							if (data.ColumnType == ColumnType.BidQuantity)
							{
								AppendRealTimeQuotes(data.RTDQuote.Instrument, time, price, volume, 0L, MarketDataType.Bid);
								AppendL2Data(data.RTDQuote.Instrument, time, 0, price, volume, SharpCharts.Base.Dom.Operation.Added, MarketDataType.Bid);
							}
							else if (data.ColumnType == ColumnType.AskQuantity)
							{
								AppendRealTimeQuotes(data.RTDQuote.Instrument, time, price, volume, 0L, MarketDataType.Ask);
								AppendL2Data(data.RTDQuote.Instrument, time, 0, price, volume, SharpCharts.Base.Dom.Operation.Added, MarketDataType.Ask); 
							}
							else if (data.ColumnType == ColumnType.TotalVolume)
							{
								AppendRealTimeQuotes(data.RTDQuote.Instrument, time, price, volume, 0L, MarketDataType.Last);
								AppendRealTimeQuotes(data.RTDQuote.Instrument, time, 0.0d, (int)array.GetValue(1, i), 0L, MarketDataType.TotalVolume);
							}
						}
					}
				}
			}
			
			
		}
		
		private int heartbeatInterval = -1;
		
		int NestTrader.IRTDUpdateEvent.HeartbeatInterval
		{
			get { return this.heartbeatInterval; }
			set { this.heartbeatInterval = value; }
		}
		
		
		
		#endregion

    }
}